/*
 *      SCCS:  @(#)sigsafe.c	1.2 (96/09/30) 
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1996 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * A copy of the end-user licence agreement is contained in the file
 * Licence which accompanies this distribution.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)sigsafe.c	1.2 (96/09/30) TETware release 3.1";
#endif

/************************************************************************

SCCS:   	@(#)sigsafe.c	1.2 96/09/30 TETware release 3.1
NAME:		sigblock.c
PRODUCT:	TETware
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	September 1996

DESCRIPTION:
	functions for blocking all blockable signals during critical
	sections of code.

MODIFICATIONS:

************************************************************************/

#include <stdio.h>
#include <signal.h>
#include "dtmac.h"
#include "dtthr.h"
#include "error.h"
#include "sigsafe.h"

#ifdef NEEDsrcFile
static char srcFile[] = __FILE__;	/* file name for error reporting */
#endif


#ifdef _WIN32

   /* macros for manipulating bits in an array of longs */
#  define BITSPERLONG	(sizeof (long) * 8)
#  define NEEDELEM(n)	(((n) + (BITSPERLONG - 1)) / BITSPERLONG)
#  define BITSET(n, a)	(a[(n) / BITSPERLONG] |= (1 << ((n) % BITSPERLONG)))
#  define BITCLR(n, a)	(a[(n) / BITSPERLONG] &= ~(1 << ((n) % BITSPERLONG)))
#  define ISSET(n, a)	(a[(n) / BITSPERLONG] & (1 << ((n) % BITSPERLONG)))

   /* signals caught by tet_sigsafe_trap() */
   static long caught_sigs[NEEDELEM(NSIG)];

   /* static function declarations */
   static void tet_sigsafe_trap PROTOLIST((int));

#endif /* _WIN32 */



#ifndef _WIN32
extern sigset_t tet_blockable_sigs;
#endif

static int init_done = 0;

/*
** tet_init_blockable_sigs() - initialise set of blockable signals
*/

void tet_init_blockable_sigs()
{
#ifdef _WIN32

	/* nothing */

#else /* _WIN32 */

	/* start with full set, and then remove signals that
	   should not be blocked */

	(void) sigfillset(&tet_blockable_sigs);

	/* the system won't allow these to be blocked */
	(void) sigdelset(&tet_blockable_sigs, SIGKILL);
	(void) sigdelset(&tet_blockable_sigs, SIGSTOP);

	/* blocking this could give problems, and it is unlikely
	   anyone would longjmp out of a SIGCHLD handler (no TET
	   code does), so it should be safe to leave it unblocked */
	(void) sigdelset(&tet_blockable_sigs, SIGCHLD);

	/* hardware signals that give undefined behaviour if blocked */
	(void) sigdelset(&tet_blockable_sigs, SIGSEGV);
	(void) sigdelset(&tet_blockable_sigs, SIGILL);
	(void) sigdelset(&tet_blockable_sigs, SIGFPE);
#  ifdef SIGBUS
	(void) sigdelset(&tet_blockable_sigs, SIGBUS);
#  endif

#endif /* _WIN32 */

	init_done = 1;
}


#ifdef _WIN32

/*
** tet_sigsafe_start() - emulate signal blocking at start of critical code
** section on WIN32 (which does not have signal masks)
**
** The list of previous signal dispositions is returned via the pointer
** argument.  This should be passed to tet_sigsafe_end().
**
** Return value is 0 for success, -1 on error.
*/

int tet_sigsafe_start(listp)
void (**listp) PROTOLIST((int));
{
	int sig;
	void (*func) PROTOLIST((int));

	ASSERT(init_done);

	/*
	** for each signal that it is safe to catch:
	**	if the signal is not currently being ignored, remember it
	**	in the user-supplied array and install a handler for use
	**	during the critical code section
	*/
	for (sig = 1; sig < NSIG; sig++) {
		BITCLR(sig, caught_sigs);
		switch (sig) {
		case SIGINT:
		case SIGTERM:
		case SIGABRT:
		case SIGBREAK:
			if ((func = signal(sig, SIG_IGN)) == SIG_ERR)
				func = SIG_IGN;
			if (func != SIG_IGN)
				(void) signal(sig, tet_sigsafe_trap);
			*(listp + sig - 1) = func;
			break;
		default:
			*(listp + sig - 1) = SIG_IGN;
			break;
		}
	}

	return(0);
}

#else /* _WIN32 */

/*
** tet_sigsafe_start() - block signals at start of critical code section
**
** The old signal mask is returned via the pointer argument.  This should
** be passed to tet_sigsafe_end().
**
** Return value is 0 for success, -1 on error.
*/

int tet_sigsafe_start(oldset)
sigset_t *oldset;
{
	ASSERT(init_done);

#ifdef TET_THREADS
	return thr_sigsetmask(SIG_BLOCK, &tet_blockable_sigs, oldset);
#else
	return sigprocmask(SIG_BLOCK, &tet_blockable_sigs, oldset);
#endif
}

#endif /* _WIN32 */



#ifdef _WIN32

/*
** tet_sigsafe_end() - unblock signals at the end of a critical code section
**
** arg is a pointer to an opaque object provided by the caller which has
** been filled in by a previous tet_sigsafe_start() call
**
** Return value is 0 for success, -1 on error.
*/

int tet_sigsafe_end(listp)
void (**listp) PROTOLIST((int));
{
	int sig;
	void (*func) PROTOLIST((int));

	/*
	** for each signal that was not previously being ignored,
	** restore the previous handler
	*/
	for (sig = 1; sig < NSIG; sig++)
		if ((func = *(listp + sig - 1)) != SIG_IGN)
			(void) signal(sig, func);

	/* resend signals which arrived during the critical code section */
	for (sig = 1; sig < NSIG; sig++)
		if (ISSET(sig, caught_sigs))
			(void) raise(sig);

	return(0);
}

#else /* _WIN32 */

/*
** tet_sigsafe_end() - restore signal mask at end of critical code section
**
** Return value is 0 for success, -1 on error.
*/

int tet_sigsafe_end(oldset)
sigset_t *oldset;
{
#ifdef TET_THREADS
	return thr_sigsetmask(SIG_SETMASK, oldset, (sigset_t *)0);
#else
	return sigprocmask(SIG_SETMASK, oldset, (sigset_t *)0);
#endif
}

#endif /* _WIN32 */



#ifdef _WIN32

/*
**	tet_sigsafe_trap() - signal handler for use during a critical
**		code section (i.e., code between tet_sigsafe_start() and
**		tet_sigsafe_end())
*/

static void tet_sigsafe_trap(sig)
int sig;
{
	ASSERT(sig > 0 && sig < NSIG);

	/* reset the handler */
	(void) signal(sig, tet_sigsafe_trap);

	/*
	** remember the signal so that the action stored
	** by tet_sigsafe_start() can be taken when tet_sigsafe_end()
	** is called
	*/
	BITSET(sig, caught_sigs);
}

#endif /* _WIN32 */


